"""Config models for summer / winter rules."""
import HABApp
import pydantic

import habapp_rules.core.pydantic_base


class SummerWinterItems(habapp_rules.core.pydantic_base.ItemBase):
	"""Items for summer/winter detection."""
	outside_temperature: HABApp.openhab.items.NumberItem = pydantic.Field(..., description="outside temperature item")
	summer: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="summer item")
	last_check: HABApp.openhab.items.DatetimeItem | None = pydantic.Field(None, description="last check item")


class SummerWinterParameter(habapp_rules.core.pydantic_base.ParameterBase):
	"""Parameter for summer/winter detection."""
	persistence_service: str | None = pydantic.Field(None, description="name of persistence service")
	days: int = pydantic.Field(5, description="number of days in the past which will be used to check if it is summer")
	temperature_threshold: float = pydantic.Field(16, description="threshold weighted temperature for summer")


class SummerWinterConfig(habapp_rules.core.pydantic_base.ConfigBase):
	"""Config for summer/winter detection."""
	items: SummerWinterItems = pydantic.Field(..., description="items for summer/winter state")
	parameter: SummerWinterParameter = pydantic.Field(SummerWinterParameter(), description="parameter for summer/winter")



"""Module for filter functions / rules."""
import logging

import HABApp

import habapp_rules.common.config.filter
import habapp_rules.core.logger

LOGGER = logging.getLogger(__name__)


class ExponentialFilter(HABApp.Rule):
	"""Rules class to apply a exponential filter to a number value.

	# Items:
	Number    BrightnessValue                       "Brightness Value [%d]"                         {channel="..."}
	Number    BrightnessFiltered                    "Brightness filtered [%d]"
	Number    BrightnessFilteredInstantIncrease     "Brightness filtered instant increase [%d]"

	# Config
	config = habapp_rules.common.config.filter.ExponentialFilterConfig(
		items = habapp_rules.common.config.filter.ExponentialFilterItems(
			raw = "BrightnessValue",
			filtered = "BrightnessFiltered"
		),
		parameter = habapp_rules.common.config.filter.ExponentialFilterParameter(  # filter constant 1 minute
			tau = 60
		)
	)

	config2 = habapp_rules.common.config.filter.ExponentialFilterConfig(
		items = habapp_rules.common.config.filter.ExponentialFilterItems(
			raw = "BrightnessValue",
			filtered = "BrightnessFilteredInstantIncrease"
		),
		parameter = habapp_rules.common.config.filter.ExponentialFilterParameter(   # filter constant 10 minutes + instant increase
			tau = 600,
			instant_increase = True
		)
	)

	# Rule init:
	habapp_rules.common.filter.ExponentialFilter(config)  # filter constant 1 minute
    habapp_rules.common.filter.ExponentialFilter(config2)  # filter constant 10 minutes + instant increase
	"""

	def __init__(self, config: habapp_rules.common.config.filter.ExponentialFilterConfig):
		"""Init exponential filter rule.

		:param config: Config for exponential filter
		"""
		HABApp.Rule.__init__(self)
		self._config = config

		self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, self.rule_name)

		self._previous_value = self._config.items.raw.value

		sample_time = self._config.parameter.tau / 5  # fifth part of the filter time constant
		self._alpha = 0.2  # always 0.2 since we always have the fifth part of the filter time constant
		self.run.every(None, sample_time, self._cb_cyclic_calculate_and_update_output)

		if self._config.parameter.instant_increase or self._config.parameter.instant_decrease:
			self._config.items.raw.listen_event(self._cb_item_raw, HABApp.openhab.events.ItemStateChangedEventFilter())

		self._instance_logger.debug(f"Successfully created exponential filter for item {self._config.items.raw.name}")

	def _cb_cyclic_calculate_and_update_output(self) -> None:
		"""Calculate the new filter output and update the filtered item. This must be called cyclic"""
		new_value = self._config.items.raw.value

		if any(not isinstance(value, (int, float)) for value in (self._previous_value, new_value)):
			self._instance_logger.warning(f"New or previous value is not a number: new_value: {new_value} | previous_value: {self._previous_value}")
			return

		self._send_output(filtered_value := self._alpha * new_value + (1 - self._alpha) * self._previous_value)
		self._previous_value = filtered_value

	def _cb_item_raw(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback which is called if the value of the raw item changed.

		:param event: event which triggered this event
		"""
		if self._previous_value is None or self._config.parameter.instant_increase and event.value > self._previous_value or self._config.parameter.instant_decrease and event.value < self._previous_value:
			self._send_output(event.value)
			self._previous_value = event.value

	def _send_output(self, new_value: float) -> None:
		"""Send output to the OpenHAB item.

		:param new_value: new value which should be sent
		"""
		self._config.items.filtered.oh_send_command(new_value)



"""Config models for logic rules."""
import typing

import HABApp
import pydantic

import habapp_rules.core.pydantic_base


class BinaryLogicItems(habapp_rules.core.pydantic_base.ItemBase):
	"""Items for binary logic."""
	inputs: list[HABApp.openhab.items.SwitchItem | HABApp.openhab.items.ContactItem] = pydantic.Field(..., description="List of input items (must be either Switch or Contact and all have to match to output_item)")
	output: HABApp.openhab.items.SwitchItem | HABApp.openhab.items.ContactItem = pydantic.Field(..., description="Output item")

	@pydantic.model_validator(mode="after")
	def validate_items(self) -> typing.Self:
		"""Validate if all items are of the same type.

		:return: validated model
		:raises ValueError: if not all items are of the same type
		"""
		for item in self.inputs:
			if not isinstance(item, type(self.output)):
				raise ValueError(f"Item '{item.name}' must have the same type like the output item. Expected: {type(self.output)} | actual : {type(item)}")
		return self


class BinaryLogicConfig(habapp_rules.core.pydantic_base.ConfigBase):
	"""Config for binary logic."""
	items: BinaryLogicItems = pydantic.Field(..., description="Items for binary logic")
	parameter: None = None


class NumericLogicItems(BinaryLogicItems):
	"""Items for numeric logic."""
	inputs: list[HABApp.openhab.items.NumberItem | HABApp.openhab.items.DimmerItem] = pydantic.Field(..., description="List of input items (must be either Number or Dimmer and all have to match to output_item)")
	output: HABApp.openhab.items.NumberItem | HABApp.openhab.items.DimmerItem = pydantic.Field(..., description="Output item")


class NumericLogicParameter(habapp_rules.core.pydantic_base.ParameterBase):
	"""Parameter for numeric logic."""
	ignore_old_values_time: int | None = pydantic.Field(None, description="ignores values which are older than the given time in seconds. If None, all values will be taken")


class NumericLogicConfig(habapp_rules.core.pydantic_base.ConfigBase):
	"""Config for numeric logic."""
	items: NumericLogicItems = pydantic.Field(..., description="Items for numeric logic")
	parameter: NumericLogicParameter = pydantic.Field(NumericLogicParameter(), description="Parameter for numeric logic")


class InvertValueItems(habapp_rules.core.pydantic_base.ItemBase):
	"""Items for invert value."""
	input: HABApp.openhab.items.NumberItem = pydantic.Field(..., description="Input item")
	output: HABApp.openhab.items.NumberItem = pydantic.Field(..., description="Output item")


class InvertValueParameter(habapp_rules.core.pydantic_base.ParameterBase):
	"""Parameter for invert value."""
	only_positive: bool = pydantic.Field(False, description="if true, only positive values will be set to output item")
	only_negative: bool = pydantic.Field(False, description="if true, only negative values will be set to output item")


class InvertValueConfig(habapp_rules.core.pydantic_base.ConfigBase):
	"""Config for invert value."""
	items: InvertValueItems = pydantic.Field(..., description="Items for invert value")
	parameter: InvertValueParameter = pydantic.Field(InvertValueParameter(), description="Parameter for invert value")



"""Rule to set/unset sleep state."""
import datetime
import logging

import HABApp.openhab.definitions
import HABApp.openhab.events
import HABApp.openhab.items
import HABApp.util

import habapp_rules.core.helper
import habapp_rules.core.logger
import habapp_rules.core.state_machine_rule
import habapp_rules.system.config.sleep

LOGGER = logging.getLogger(__name__)


# pylint: disable=no-member
class Sleep(habapp_rules.core.state_machine_rule.StateMachineRule):
	"""Rules class to manage sleep state.

	Example OpenHAB configuration:
	# KNX-things:
	Thing device T00_99_OpenHab_Sleep "KNX OpenHAB Sleep"{
        Type switch             : sleep             "Sleep Request"             [ ga="0/2/30"]
        Type switch-control     : sleep_RM          "Sleep RM"                  [ ga="0/2/31"]

        Type switch             : sleep_lock        "Sleep Lock Request"        [ ga="0/2/32"]
        Type switch-control     : sleep_lock_RM     "Sleep Lock RM"             [ ga="0/2/33"]

        Type string-control     : sleep_text        "Sleep Text"                [ ga="16.000:0/2/34"]
    }

    # Items:
    Switch    I01_02_Sleep              "Sleep [%s]"                <moon>     {channel="knx:device:bridge:T00_99_OpenHab_Sleep:sleep_RM"}
	Switch    I01_02_Sleep_req          "Sleep request"             <moon>     {channel="knx:device:bridge:T00_99_OpenHab_Sleep:sleep"}
	String    I01_02_Sleep_text         "Text for display [%s]"                {channel="knx:device:bridge:T00_99_OpenHab_Sleep:sleep_text"}
	Switch    I01_02_Sleep_lock         "Lock [%s]"                 <lock>     {channel="knx:device:bridge:T00_99_OpenHab_Sleep:sleep_lock_RM"}
	Switch    I01_02_Sleep_lock_req     "Lock request"              <lock>     {channel="knx:device:bridge:T00_99_OpenHab_Sleep:sleep_lock"}
	String    I01_02_Sleep_State        "State [%s]"                <state>

	# Config:
	config = habapp_rules.system.config.sleep.SleepConfig(
		items=habapp_rules.system.config.sleep.SleepItems(
			sleep="I01_02_Sleep",
			sleep_req="I01_02_Sleep_req",
			state="I01_02_Sleep_State",
			lock="I01_02_Sleep_lock",
			lock_req="I01_02_Sleep_lock_req",
			display_text="I01_02_Sleep_text"
		)
	)

	# Rule init:
	habapp_rules.system.sleep.Sleep(config)
	"""

	states = [
		{"name": "awake"},
		{"name": "pre_sleeping", "timeout": 3, "on_timeout": "pre_sleeping_timeout"},
		{"name": "sleeping"},
		{"name": "post_sleeping", "timeout": 3, "on_timeout": "post_sleeping_timeout"},
		{"name": "locked"},
	]

	trans = [
		{"trigger": "start_sleeping", "source": "awake", "dest": "pre_sleeping"},
		{"trigger": "pre_sleeping_timeout", "source": "pre_sleeping", "dest": "sleeping"},
		{"trigger": "end_sleeping", "source": "sleeping", "dest": "post_sleeping"},
		{"trigger": "end_sleeping", "source": "pre_sleeping", "dest": "awake", "unless": "lock_request_active"},
		{"trigger": "end_sleeping", "source": "pre_sleeping", "dest": "locked", "conditions": "lock_request_active"},
		{"trigger": "post_sleeping_timeout", "source": "post_sleeping", "dest": "awake", "unless": "lock_request_active"},
		{"trigger": "post_sleeping_timeout", "source": "post_sleeping", "dest": "locked", "conditions": "lock_request_active"},
		{"trigger": "set_lock", "source": "awake", "dest": "locked"},
		{"trigger": "release_lock", "source": "locked", "dest": "awake"}
	]

	def __init__(self, config: habapp_rules.system.config.sleep.SleepConfig) -> None:
		"""Init of Sleep object.

		:param config: config for sleeping state
		"""
		self._config = config
		habapp_rules.core.state_machine_rule.StateMachineRule.__init__(self, config.items.state)
		self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, config.items.sleep.name)

		# init attributes
		self._sleep_request_active = config.items.sleep_request.is_on()
		self._lock_request_active = config.items.lock_request.is_on() if config.items.lock_request is not None else False

		# init state machine
		self.state_machine = habapp_rules.core.state_machine_rule.StateMachineWithTimeout(
			model=self,
			states=self.states,
			transitions=self.trans,
			ignore_invalid_triggers=True,
			after_state_change="_update_openhab_state")
		self._set_initial_state()

		self._update_openhab_state()

		# add callbacks
		config.items.sleep_request.listen_event(self._cb_sleep_request, HABApp.openhab.events.ItemStateChangedEventFilter())
		if config.items.lock_request is not None:
			config.items.lock_request.listen_event(self._cb_lock_request, HABApp.openhab.events.ItemStateChangedEventFilter())

		self._instance_logger.debug(super().get_initial_log_message())

	def _get_initial_state(self, default_value: str = "awake") -> str:
		"""Get initial state of state machine.

		:param default_value: default / initial state
		:return: return correct state if it could be detected, if not return default value
		"""
		sleep_req = self._config.items.sleep_request.is_on() if self._config.items.sleep_request.value is not None else None
		lock_req = self._config.items.lock_request.is_on() if self._config.items.lock_request is not None and self._config.items.lock_request.value is not None else None

		if sleep_req:
			return "sleeping"
		if lock_req:
			return "locked"
		if sleep_req is False:
			return "awake"

		return default_value

	@property
	def sleep_request_active(self) -> bool:
		"""Check if a sleep request is active

		:return: return true if lock request is active
		"""
		return self._sleep_request_active

	@property
	def lock_request_active(self) -> bool:
		"""Check if a lock request is active

		:return: return true if lock request is active
		"""
		return self._lock_request_active

	def _update_openhab_state(self):
		"""Extend _update_openhab state of base class to also update other OpenHAB items."""
		super()._update_openhab_state()

		# update sleep state
		if self.state in {"pre_sleeping", "sleeping"}:
			habapp_rules.core.helper.send_if_different(self._config.items.sleep, "ON")
		else:
			habapp_rules.core.helper.send_if_different(self._config.items.sleep, "OFF")

		# update lock state
		self.__update_lock_state()

		# update display text
		if self._config.items.display_text is not None:
			self._config.items.display_text.oh_send_command(self.__get_display_text())

	def __get_display_text(self) -> str:
		"""Get Text for displays.

		:return: display text
		"""
		if self.state == "awake":
			return "Schlafen"
		if self.state == "pre_sleeping":
			return "Guten Schlaf"
		if self.state == "sleeping":
			return "Aufstehen"
		if self.state == "post_sleeping":
			return "Guten Morgen"
		if self.state == "locked":
			return "Gesperrt"
		return ""

	def __update_lock_state(self):
		"""Update the return lock state value of OpenHAB item."""
		if self._config.items.lock is not None:
			if self.state in {"pre_sleeping", "post_sleeping", "locked"}:
				habapp_rules.core.helper.send_if_different(self._config.items.lock, "ON")
			else:
				habapp_rules.core.helper.send_if_different(self._config.items.lock, "OFF")

	def _cb_sleep_request(self, event: HABApp.openhab.events.ItemStateChangedEvent):
		"""Callback, which is called if sleep request item changed state.

		:param event: Item state change event of sleep_request item
		"""
		if event.value == "ON" and self.state == "awake":
			self._instance_logger.debug("Start sleeping through sleep switch")
			self._sleep_request_active = True
			self.start_sleeping()
		elif event.value == "ON" and self.state == "locked":
			self._sleep_request_active = False
			self._config.items.sleep_request.oh_send_command("OFF")
		elif event.value == "OFF" and self.state in {"sleeping", "pre_sleeping"}:
			self._instance_logger.debug("End sleeping through sleep switch")
			self._sleep_request_active = True
			self.end_sleeping()

	def _cb_lock_request(self, event: HABApp.openhab.events.ItemStateChangedEvent):
		"""Callback, which is called if lock request item changed state.

		:param event: Item state change event of sleep_request item
		"""
		self._lock_request_active = event.value == "ON"

		if self.state == "awake" and event.value == "ON":
			self.set_lock()
		elif self.state == "locked" and event.value == "OFF":
			self.release_lock()
		else:
			self.__update_lock_state()


class LinkSleep(HABApp.Rule):
	"""Link sleep items depending on current time"""

	def __init__(self, config: habapp_rules.system.config.sleep.LinkSleepConfig) -> None:
		"""Init rule.

		:param config: Config for linking sleep rules
		"""
		self._config = config
		HABApp.Rule.__init__(self)
		self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, self.rule_name)

		config.items.sleep_master.listen_event(self._cb_master, HABApp.openhab.events.ItemStateChangedEventFilter())

		if config.items.link_active_feedback is not None:
			self.run.on_every_day(config.parameter.link_time_start, self._set_link_active_feedback, target_state="ON")
			self.run.on_every_day(config.parameter.link_time_end, self._set_link_active_feedback, target_state="OFF")
			self.run.soon(self._set_link_active_feedback, target_state=self._check_time_in_window())

	def _check_time_in_window(self) -> bool:
		"""Check if current time is in the active time window

		:return: True if current time is in time the active time window
		"""
		now = datetime.datetime.now().time()

		if self._config.parameter.link_time_start <= self._config.parameter.link_time_end:
			return self._config.parameter.link_time_start <= now <= self._config.parameter.link_time_end
		# cross midnight
		return self._config.parameter.link_time_start <= now or now <= self._config.parameter.link_time_end

	def _cb_master(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback which is triggered if the state of the master item changes

		:param event: state change event
		"""
		if not self._check_time_in_window():
			return

		self._instance_logger.debug(f"Set request of all linked sleep states of {self._config.items.sleep_master.name}")
		for itm in self._config.items.sleep_request_slaves:
			habapp_rules.core.helper.send_if_different(itm, event.value)

	def _set_link_active_feedback(self, target_state: str) -> None:
		"""Set feedback for link is active.

		:param target_state: Target state which should be set ["ON" / "OFF"]
		"""
		self._config.items.link_active_feedback.oh_send_command(target_state)



"""Module for sending the monthly energy consumption."""
import datetime
import logging
import pathlib
import tempfile

import HABApp
import HABApp.core.internals
import dateutil.relativedelta
import jinja2
import multi_notifier.connectors.connector_mail

import habapp_rules.__version__
import habapp_rules.core.exceptions
import habapp_rules.core.logger
import habapp_rules.energy.config.monthly_report
import habapp_rules.energy.donut_chart

LOGGER = logging.getLogger(__name__)

MONTH_MAPPING = {
	1: "Januar",
	2: "Februar",
	3: "März",
	4: "April",
	5: "Mai",
	6: "Juni",
	7: "Juli",
	8: "August",
	9: "September",
	10: "Oktober",
	11: "November",
	12: "Dezember"
}


def _get_previous_month_name() -> str:
	"""Get name of the previous month

	:return: name of current month

	if other languages are required, the global dict must be replaced
	"""
	today = datetime.date.today()
	last_month = today.replace(day=1) - datetime.timedelta(days=1)

	return MONTH_MAPPING[last_month.month]


def _get_next_trigger() -> datetime.datetime:
	"""Get next trigger time (always first day of month at midnight)

	:return: next trigger time
	"""
	return datetime.datetime.now().replace(day=1, hour=0, minute=0, second=0, microsecond=0) + dateutil.relativedelta.relativedelta(months=1)


class MonthlyReport(HABApp.Rule):
	"""Rule for sending the monthly energy consumption.

	# Config
	config = habapp_rules.energy.config.monthly_report.MonthlyReportConfig(
		items=habapp_rules.energy.config.monthly_report.MonthlyReportItems(
			energy_sum="Total Energy"
		),
		parameter=habapp_rules.energy.config.monthly_report.MonthlyReportParameter(
			known_energy_shares=[
				habapp_rules.energy.config.monthly_report.EnergyShare("Dishwasher_Energy", "Dishwasher"),
				habapp_rules.energy.config.monthly_report.EnergyShare("Light", "Light")
			],
			config_mail=multi_notifier.connectors.connector_mail.MailConfig(
				user="sender@test.de",
				password="fancy_password",
				smtp_host="smtp.test.de",
				smtp_port=587
			),
			recipients=["test@test.de"],
		)
	)

	# Rule init
	habapp_rules.energy.monthly_report.MonthlyReport("Total_Energy", known_energy_share, "Group_RRD4J", config_mail, "test@test.de")
	"""

	def __init__(self, config: habapp_rules.energy.config.monthly_report.MonthlyReportConfig) -> None:
		"""Initialize the rule.

		:param config: config for the monthly energy report rule
		:raises habapp_rules.core.exceptions.HabAppRulesConfigurationException: if config is not valid
		"""
		self._config = config
		HABApp.Rule.__init__(self)
		self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, config.items.energy_sum.name)
		self._mail = multi_notifier.connectors.connector_mail.Mail(config.parameter.config_mail)

		self._mail = multi_notifier.connectors.connector_mail.Mail(config.parameter.config_mail)

		if config.parameter.persistence_group_name is not None:
			# check if all energy items are in the given persistence group
			items_to_check = [config.items.energy_sum] + [share.energy_item for share in config.parameter.known_energy_shares]
			not_in_persistence_group = [item.name for item in items_to_check if config.parameter.persistence_group_name not in item.groups]
			if not_in_persistence_group:
				raise habapp_rules.core.exceptions.HabAppRulesConfigurationException(f"The following OpenHAB items are not in the persistence group '{config.parameter.persistence_group_name}': {not_in_persistence_group}")

		self.run.at(next_trigger_time := _get_next_trigger(), self._cb_send_energy)
		if config.parameter.debug:
			self._instance_logger.warning("Debug mode is active!")
			self.run.soon(self._cb_send_energy)
		self._instance_logger.info(f"Successfully initiated monthly consumption rule for {config.items.energy_sum}. Triggered first execution to {next_trigger_time.isoformat()}")

	def _get_historic_value(self, item: HABApp.openhab.items.NumberItem, start_time: datetime.datetime) -> float:
		"""Get historic value of given Number item

		:param item: item instance
		:param start_time: start time to search for the interested value
		:return: historic value of the item
		"""
		historic = item.get_persistence_data(start_time=start_time, end_time=start_time + datetime.timedelta(hours=1)).data
		if not historic:
			self._instance_logger.warning(f"Could not get value of item '{item.name}' of time = {start_time}")
			return 0

		return next(iter(historic.values()))

	# pylint: disable=wrong-spelling-in-docstring
	def _create_html(self, energy_sum_month: float) -> str:
		"""Create html which will be sent by the mail

		:param energy_sum_month: sum value for the current month
		:return: html with replaced values

		The template was created by https://app.bootstrapemail.com/editor/documents with the following input:

		<html>
		  <head>
		    <meta http-equiv="Content-Type" content="text/html; charset=utf-8" />
		    <style>
		    </style>
		  </head>
		  <body class="bg-light">
		    <div class="container">
		      <div class="card my-10">
		        <div class="card-body">
		          <h1 class="h3 mb-2">Strom Verbrauch</h1>
		          <h5 class="text-teal-700">von Februar</h5>
		          <hr>
		          <div class="space-y-3">
		            <p class="text-gray-700">Aktueller Zählerstand: <b>7000 kWh</b>.</p>
		            <p class="text-gray-700">Hier die Details:</p>
		            <p><img src="https://www.datylon.com/hubfs/Datylon%20Website2020/Datylon%20Chart%20library/Chart%20pages/Pie%20Chart/datylon-chart-library-pie-chart-intro-example.svg" alt="Italian Trulli" align="left">
		            </p>
		          </div>
		          <hr>
		           <p style="font-size: 0.6em">Generated with habapp_rules version = 20.0.3</p>
		        </div>
		      </div>
		    </div>
		  </body>
		</html>
		"""
		with (pathlib.Path(__file__).parent / "monthly_report_template.html").open(encoding="utf-8") as template_file:
			html_template = template_file.read()

		return jinja2.Template(html_template).render(
			month=_get_previous_month_name(),
			energy_now=f"{self._config.items.energy_sum.value:.1f}",
			energy_last_month=f"{energy_sum_month:.1f}",
			habapp_version=habapp_rules.__version__.__version__,
			chart="{{ chart }}"  # this is needed to not replace the chart from the mail-template
		)

	def _cb_send_energy(self) -> None:
		"""Send the mail with the energy consumption of the last month"""
		self._instance_logger.debug("Send energy consumption was triggered.")
		# get values
		now = datetime.datetime.now()
		last_month = now - dateutil.relativedelta.relativedelta(months=1)

		energy_sum_month = self._config.items.energy_sum.value - self._get_historic_value(self._config.items.energy_sum, last_month)
		for share in self._config.parameter.known_energy_shares:
			share.monthly_power = share.energy_item.value - self._get_historic_value(share.energy_item, last_month)

		energy_unknown = energy_sum_month - sum(share.monthly_power for share in self._config.parameter.known_energy_shares)

		with tempfile.TemporaryDirectory() as temp_dir_name:
			# create plot
			labels = [share.chart_name for share in self._config.parameter.known_energy_shares] + ["Rest"]
			values = [share.monthly_power for share in self._config.parameter.known_energy_shares] + [energy_unknown]
			chart_path = pathlib.Path(temp_dir_name) / "chart.png"
			habapp_rules.energy.donut_chart.create_chart(labels, values, chart_path)

			# get html
			html = self._create_html(energy_sum_month)

			# send mail
			self._mail.send_message(self._config.parameter.recipients, html, f"Stromverbrauch {_get_previous_month_name()}", images={"chart": str(chart_path)})

		self.run.at(next_trigger_time := _get_next_trigger(), self._cb_send_energy)
		self._instance_logger.info(f"Successfully sent energy consumption mail to {self._config.parameter.recipients}. Scheduled the next trigger time to {next_trigger_time.isoformat()}")



"""Config models for irrigation rules."""
import typing

import HABApp.openhab.items
import pydantic

import habapp_rules.core.exceptions
import habapp_rules.core.pydantic_base


class IrrigationItems(habapp_rules.core.pydantic_base.ItemBase):
	"""Items for irrigation rules."""
	valve: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="valve item which will be switched")
	active: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="item to activate the rule")
	hour: HABApp.openhab.items.NumberItem = pydantic.Field(..., description="start hour")
	minute: HABApp.openhab.items.NumberItem = pydantic.Field(..., description="start minute")
	duration: HABApp.openhab.items.NumberItem = pydantic.Field(..., description="duration in minutes")
	repetitions: HABApp.openhab.items.NumberItem | None = pydantic.Field(None, description="number of repetitions")
	brake: HABApp.openhab.items.NumberItem | None = pydantic.Field(None, description="time in minutes between repetitions")

	@pydantic.model_validator(mode="after")
	def validate_model(self) -> typing.Self:
		"""Validate model

		:return: validated model
		:raises AssertionError: if 'repetitions' and 'brake' are not set together
		"""
		if (self.repetitions is None) != (self.brake is None):
			raise AssertionError("If repeats item is given, also the brake item must be given!")

		return self


class IrrigationConfig(habapp_rules.core.pydantic_base.ConfigBase):
	"""Config for irrigation actors."""
	items: IrrigationItems
	parameter: None =  None



"""Config models for motion rules."""
import HABApp
import pydantic

import habapp_rules.core.pydantic_base


class MotionItems(habapp_rules.core.pydantic_base.ItemBase):
	"""Items for motion."""
	motion_raw: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="unfiltered motion item")
	motion_filtered: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="filtered motion item")
	brightness: HABApp.openhab.items.NumberItem | None = pydantic.Field(None, description="brightness item")
	brightness_threshold: HABApp.openhab.items.NumberItem | None = pydantic.Field(None, description="brightness threshold item")
	lock: HABApp.openhab.items.SwitchItem | None = pydantic.Field(None, description="lock item")
	sleep_state: HABApp.openhab.items.StringItem | None = pydantic.Field(None, description="sleep state item")
	state: HABApp.openhab.items.StringItem = pydantic.Field(..., description="state item")


class MotionParameter(habapp_rules.core.pydantic_base.ParameterBase):
	"""Parameter for motion."""
	extended_motion_time: int = pydantic.Field(5, description="extended motion time in seconds")
	brightness_threshold: float | None = pydantic.Field(None, description="brightness threshold value")
	post_sleep_lock_time: int = pydantic.Field(10, description="post sleep lock time in seconds")


class MotionConfig(habapp_rules.core.pydantic_base.ConfigBase):
	"""Config for motion."""
	items: MotionItems = pydantic.Field(..., description="items for motion")
	parameter: MotionParameter = pydantic.Field(MotionParameter(), description="parameter for motion")



"""Config models for HCL color rules."""
import typing

import HABApp
import pydantic

import habapp_rules.core.pydantic_base


class HclTimeItems(habapp_rules.core.pydantic_base.ItemBase):
	"""Items for HCL color which depends on time"""
	color: HABApp.openhab.items.NumberItem = pydantic.Field(..., description="HCL color which will be set by the HCL rule")
	manual: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="switch item to disable all automatic functions")
	sleep_state: HABApp.openhab.items.StringItem | None = pydantic.Field(None, description="sleep state item")
	focus: HABApp.openhab.items.SwitchItem | None = pydantic.Field(None, description="focus state item")
	switch_on: HABApp.openhab.items.SwitchItem | None = pydantic.Field(None, description="switch item which triggers a color update if switched on")
	state: HABApp.openhab.items.StringItem | None = pydantic.Field(..., description="state item for storing the current state")


class HclElevationItems(HclTimeItems):
	"""Items for HCL color which depends on sun elevation"""
	elevation: HABApp.openhab.items.NumberItem = pydantic.Field(..., description="sun elevation")


class HclElevationParameter(habapp_rules.core.pydantic_base.ParameterBase):
	"""Parameter for HCL color"""
	color_map: list[tuple[int, int]] = pydantic.Field([(-15, 3900), (0, 4500), (5, 5500), (15, 6500)], description="Color mapping. The first value is the sun elevation, the second is the HCL color")
	hand_timeout: int = pydantic.Field(18_000, description="hand timeout. After this time the HCL light rule will fall back to auto mode", gt=0)  # 5 hours
	sleep_color: float = pydantic.Field(2500, description="color if sleeping is active", gt=0)
	post_sleep_timeout: int = pydantic.Field(1, description="time after sleeping was active where the sleeping color will be set", gt=0)
	focus_color: float = pydantic.Field(6000, description="color if focus is active", gt=0)

	@pydantic.model_validator(mode="after")
	def validate_model(self) -> typing.Self:
		"""Sort color map

		:return: model with sorted color map
		"""
		self.color_map = sorted(self.color_map, key=lambda x: x[0])
		return self


_DEFAULT_TIME_MAP = [
	(0, 2200),
	(4, 2200),
	(5, 3200),
	(6, 3940),
	(8, 5000),
	(12, 7000),
	(19, 7000),
	(21, 5450),
	(22, 4000),
	(23, 2600)
]


class HclTimeParameter(HclElevationParameter):
	"""Parameter for HCL color which depends on time"""
	color_map: list[tuple[int, int]] = pydantic.Field(_DEFAULT_TIME_MAP, description="Color mapping. The first value is the hour, the second is the HCL color")
	shift_weekend_holiday: bool = pydantic.Field(False, description="If this is active the color will shift on weekends and holidays for one hour")


class HclElevationConfig(habapp_rules.core.pydantic_base.ConfigBase):
	"""Config for HCL color which depends on sun elevation"""
	items: HclElevationItems = pydantic.Field(..., description="items for HCL color which depends on sun elevation")
	parameter: HclElevationParameter = pydantic.Field(HclElevationParameter(), description="parameter for HCL color which depends on sun elevation")


class HclTimeConfig(HclElevationConfig):
	"""Config for HCL color which depends on time"""
	items: HclTimeItems = pydantic.Field(..., description="items for HCL color which depends on time")
	parameter: HclTimeParameter = pydantic.Field(HclTimeParameter(), description="parameter for HCL color which depends on time")



"""Config rules for DWD rules."""
import typing

import HABApp
import pydantic

import habapp_rules.core.pydantic_base


class WindAlarmItems(habapp_rules.core.pydantic_base.ItemBase):
	"""Items for DWD wind alarm rule."""

	wind_alarm: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="item for wind alarm, which will be set to ON if wind alarm is active")
	manual: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="switch item to disable all automatic functions")
	hand_timeout: HABApp.openhab.items.NumberItem | None = pydantic.Field(None, description="item to set the hand timeout")
	state: HABApp.openhab.items.StringItem = pydantic.Field(..., description="item for storing the current state")


class WindAlarmParameter(habapp_rules.core.pydantic_base.ParameterBase):
	"""Parameter for DWD wind alarm rule."""
	hand_timeout: int | None = pydantic.Field(None, description="hand timeout in seconds or 0 for no timeout")
	dwd_item_prefix: str = pydantic.Field("I26_99_warning_", description="prefix of dwd warning names")
	number_dwd_objects: int = pydantic.Field(3, description="number of dwd objects")
	threshold_wind_speed: int = pydantic.Field(70, description="threshold for wind speed -> wind alarm will only be active if greater or equal")
	threshold_severity: int = pydantic.Field(2, description="threshold for severity -> wind alarm will only be active if greater or equal")


class WindAlarmConfig(habapp_rules.core.pydantic_base.ConfigBase):
	"""Config for DWD wind alarm rule."""
	items: WindAlarmItems = pydantic.Field(..., description="items for DWD wind alarm rule")
	parameter: WindAlarmParameter = pydantic.Field(WindAlarmParameter(), description="parameters for DWD wind alarm rule")

	@pydantic.model_validator(mode="after")
	def check_hand_timeout(self) -> typing.Self:
		"""Validate hand timeout.

		:return: validated config model
		:raises ValueError: if both 'items.hand_timeout' and 'parameter.hand_timeout' are set
		"""
		if not (self.items.hand_timeout is None) ^ (self.parameter.hand_timeout is None):  # XNOR
			raise ValueError("Either 'items.wind_alarm' or 'parameter.hand_timeout' must be set")
		return self



"""Rules for astro actions."""
import abc
import logging

import HABApp

import habapp_rules.core.helper
import habapp_rules.sensors.config.astro

LOGGER = logging.getLogger(__name__)


class _SetNightDayBase(HABApp.Rule):
	"""Base class for set night / day."""

	def __init__(self, item_target: HABApp.openhab.items.SwitchItem, item_elevation: HABApp.openhab.items.NumberItem, elevation_threshold: float) -> None:
		"""Init Rule.

		:param item_target: OpenHab item which should be set depending on the sun elevation value
		:param item_elevation: OpenHAB item of sun elevation (NumberItem)
		:param elevation_threshold: Threshold value for elevation.
		"""
		HABApp.Rule.__init__(self)

		self._item_target = item_target
		self._item_elevation = item_elevation
		self._elevation_threshold = elevation_threshold

		self._item_elevation.listen_event(self._set_night, HABApp.openhab.events.ItemStateChangedEventFilter())

		self.run.soon(self._set_night)

	def _set_night(self, _: HABApp.openhab.events.ItemStateChangedEvent | None = None) -> None:
		"""Callback which sets the state to the night item."""
		if self._item_elevation.value is None:
			return
		habapp_rules.core.helper.send_if_different(self._item_target, self._get_target_value())

	@abc.abstractmethod
	def _get_target_value(self) -> str:
		"""Get target value which should be set.

		:return: target value (ON / OFF)
		"""


class SetDay(_SetNightDayBase):
	"""Rule to set / unset day item at dusk / dawn.

	# Items:
	Switch    day                   "Day"
	Number    elevation             "Sun elevation [%s]"    <sun>     {channel="astro...}

	# Config:
	config = habapp_rules.sensors.config.astro.SetDayConfig(
		items=habapp_rules.sensors.config.astro.SetDayItems(
			day="day",
			elevation="elevation"
		),
		parameter=habapp_rules.sensors.config.astro.SetDayParameter(
			elevation_threshold=5
		)
	)

	# Rule init:
	habapp_rules.sensors.astro.SetNight(config)
	"""

	def __init__(self, config: habapp_rules.sensors.config.astro.SetDayConfig) -> None:
		"""Init Rule.

		:param config: Config for set day rule
		"""
		_SetNightDayBase.__init__(self, config.items.day, config.items.elevation, config.parameter.elevation_threshold)

	def _get_target_value(self) -> str:
		"""Get target value which should be set.

		:return: target value (ON / OFF)
		"""
		return "ON" if self._item_elevation.value > self._elevation_threshold else "OFF"


class SetNight(_SetNightDayBase):
	"""Rule to set / unset night item at dusk / dawn.

	# Items:
	Switch    night_for_shading     "Night for shading"
	Number    elevation             "Sun elevation [%s]"    <sun>     {channel="astro...}

	# Config:
	config = habapp_rules.sensors.config.astro.SetNightConfig(
		items=habapp_rules.sensors.config.astro.SetNightItems(
			night="night_for_shading",
			elevation="elevation"
		),
		parameter=habapp_rules.sensors.config.astro.SetNightParameter(
			elevation_threshold=5
		)
	)

	# Rule init:
	habapp_rules.sensors.astro.SetNight(config)
	"""

	def __init__(self, config: habapp_rules.sensors.config.astro.SetNightConfig) -> None:
		"""Init Rule.

		:param config: Config for setting night depending on sun elevation
		"""
		_SetNightDayBase.__init__(self, config.items.night, config.items.elevation, config.parameter.elevation_threshold)

	def _get_target_value(self) -> str:
		"""Get target value which should be set.

		:return: target value (ON / OFF)
		"""
		return "ON" if self._item_elevation.value < self._elevation_threshold else "OFF"



"""Config models for presence rules."""
import HABApp
import pydantic

import habapp_rules.core.pydantic_base


class PresenceItems(habapp_rules.core.pydantic_base.ItemBase):
	"""Items for presence detection."""
	presence: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="presence item")
	leaving: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="leaving item")
	outdoor_doors: list[HABApp.openhab.items.ContactItem] = pydantic.Field([], description="list of door contacts which are used to detect presence if outside door was opened")
	phones: list[HABApp.openhab.items.SwitchItem] = pydantic.Field([], description="list of phone items which are used to detect presence and leaving depending on present phones")
	state: HABApp.openhab.items.StringItem = pydantic.Field(..., description="state item")


class PresenceConfig(habapp_rules.core.pydantic_base.ConfigBase):
	"""Config for presence detection."""
	items: PresenceItems = pydantic.Field(..., description="items for presence detection")
	parameter: None = None
